//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: C++ Client Service Example
//
// Description:
//
// Implementation
//
// Date:   Wed 05/17/2007
// Revision: $Revision: 1.4 $
//
// Last Edit Date: $Date: 2007/06/20 16:21:19 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/ClientService/CHMclient.cpp,v $
//---------------------------------------------------------------------------
#include "CHMclientServicePrecomp.h"
#pragma hdrstop

#include <BIZ/BIZutils.h>
#include <SERV/SERVostream.h>
#include <SERV/SERVutils.h>
#include "CHMclient.h"
#include "CHMconstants.h"

static const CHMint32 MAX_MESSAGE_LENGTH = 200;

void CHMtransmitWithLog(CHMllpClient& Client, const CHMstring& Message, SERVostream& Log)
{
   Log << "Sending:" << newline
       << Client.MakeMessagePrintable(Message.c_str(), APP_NEW_LINE, MAX_MESSAGE_LENGTH) << newline;
   Client.TransmitMessage(Message.c_str());
   Log << newline << "Sent!" << newline;
}

void CHMpollAndSendMessage(BIZclientState& ClientState, CHMllpClient& Client, SERVostream& Log)
{
   Log << "Polling for message to send...";

   if (BIZfetchMessage(ClientState))
   {
      Log << "found id=" << ClientState.messageId().c_str() << newline;
      // We have a message to send
      BIZsetAckTimeOut(ClientState, APP_ACK_TIMEOUT);
      // Sending the message over the network always has the potential of resulting in an exception since the remote host
      // can disconnect at any time.  The application behavior is safe since it should simply time out
      // and then resend the message.
      CHMtransmitWithLog(Client, ClientState.message(), Log);
   }
   else
   {
      Log << "None" << newline;
   }
}

void CHMconnectedAction(BIZclientState& ClientState, CHMllpClient& Client, SERVostream& Log)
{
   CHMPRECONDITION(Client.IsConnected());
   switch (ClientState.currentState())
   {
   case BIZ_NEED_TO_POLL_APP:
      CHMpollAndSendMessage(ClientState, Client, Log);
      return;
      break;
   case BIZ_WAITING_FOR_ACK:
      if (BIZisTimedOut(ClientState))
      {
         Log << "Timed out waiting for ACK for " << ClientState.messageId().c_str() << newline;
         // We need to resend the message so we reset the ACK timeout
         // TODO - we're just resending the message here; other possible implementations would be to disconnect
         // before trying to resend.

         // TODO - if we get many resends in a row it would probably make sense to keep a count and then
         // send an email or pager alert.
         BIZsetAckTimeOut(ClientState, APP_ACK_TIMEOUT);
         Log << "Resending..." << newline;
         // And resend the message
         CHMtransmitWithLog(Client, ClientState.message(), Log);
      }
      return;
      break;
   };
}

void CHMreceivedAckAction(BIZclientState& ClientState, CHMllpClient& Client, const char* pAckMessage, SERVostream& Log)
{
   CHMPRECONDITION(Client.IsConnected());
   Log << "Received an ack:" << newline;
   Log << Client.MakeMessagePrintable(pAckMessage, APP_NEW_LINE, MAX_MESSAGE_LENGTH) << newline;

   switch(ClientState.currentState())
   {
   case BIZ_WAITING_FOR_ACK:
      // TODO - we'll assume that the ACK is well formed and the message control ID matches the one we were expecting.
      // We could parse that out with Chameleon if we wanted to be very paranoid.  In a synchronous exchange it should
      // be okay...
      // The following command is *VERY IMPORTANT*. It marks the message as being sent.  The BIZ library should
      // be written in a manner to never consider a message sent until this application gives confirmation.
      BIZconfimMessageSend(ClientState, Log);
      CHMpollAndSendMessage(ClientState, Client, Log);
      break;
   case BIZ_NEED_TO_POLL_APP:
      // TODO- page alert
      Log << "WARNING!  We were not expecting an ACK." << newline;
      break;
   }
}


// This event fires when the Client connects to the Server.
// The connected event is very important.  One common mistake
// that developers often make when new to socket programming
// is to try and invoke the Connect method and then immediately
// send data via the socket.  This is not possible to do until the
// socket is actually connected.  Instead,  you should invoke the
// Connect method and then wait for this event before sending
// data.
void CHMclient::OnConnected()
{
   try
   {
      SERVcout << "Client Connected." << newline << newline << flush;
      CHMconnectedAction(ClientState, *this, SERVcout);
   }
   catch (const CHMerror& Error) 
   {
      SERVcerr << "ERROR: OnConnected" << newline << Error.Description() << newline << flush;
   }
   catch (...)
   {
      SERVcerr << "ERROR: OnConnected" << newline << "Unknown exception." << newline << flush;
   }
}

// This event fires when the Client is disconnected from the Server.
void CHMclient::OnClosed()
{
   // TODO - it's very important that we can assume that SERVcout doesn't throw;
   // otherwise, we should catch all C++ exceptions here.
   SERVcout <<  "Client Disconnected." << newline << flush;
}

// This event fires when a message is received.  Usually the message received 
// is an ACK message.  If, however, you are implementing a lab query interface 
// it might be a lab result message.  For this example we are assuming it is 
// an ACK message and just printing it.
void CHMclient::OnMessageReceived(const char* pAckMessage)
{
   try
   {
      CHMreceivedAckAction(ClientState, *this, pAckMessage, SERVcout);
   }
   catch (const CHMerror& Error) 
   {
      SERVcerr << "ERROR: OnMessageReceived" << newline << Error.Description() << newline << flush;
   }
   catch (...)
   {
      SERVcerr << "ERROR: OnMessageReceived" << newline << "Unknown exception." << newline << flush;
   }
}

// This event fires when we receive a TCP/IP error during connection.
// This will always terminate the connection, so your real application will 
// need logic to determine if and when to re-connect to the other host.
void CHMclient::OnErrorReceived(const char* Description, CHMint32 Code)
{
   // TODO - catch C++ exceptions or assume SERVcout doesn't throw 
   SERVcout << Description << newline << flush;
}

// This event fires when we receive data that is not using the Lower Layer Protocol (LLP).
// LLP is the standard protocol for sending HL7 messages, and is used by Chameleon.
// See the manual for more details.
void CHMclient::OnDataIgnored(const char* Data)
{
   // TODO - we will need to catch C++ exceptions if SERVcout can throw C++ exceptions. 
   SERVcout << "Received Ignored Data:" << newline << Data << newline << flush;
}

void CHMclientConnect(CHMllpClient& Client, SERVostream& LogStream)
{
   // Connect to the remote host.
   LogStream << "Connecting to port " << APP_REMOTE_PORT << " on " << APP_REMOTE_HOST << newline;
   Client.Connect(APP_REMOTE_HOST, APP_REMOTE_PORT);
}

void CHMclient::onTimeout()
{
   try
   {
      if (!IsConnected())
      {
         CHMclientConnect(*this, SERVcout);
         return;
      }
      CHMconnectedAction(ClientState, *this, SERVcout);
   }
   catch (const CHMerror& Error) 
   {
      SERVcerr << "ERROR: OnMessageReceived" << newline << Error.Description() << newline << flush;
   }
   catch (...)
   {
      SERVcerr << "ERROR: OnMessageReceived" << newline << "Unknown exception." << newline << flush;
   }
}
