#ifndef __SERV_BINARY_FILE_H__
#define __SERV_BINARY_FILE_H__
//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVbinaryFile
//
// Description:
//
// This object represents a binary file.  There are a couple of
// implementations under the hood.  One is using the unbuffered
// IO - open et al. which may be optimal for many applications.
// The operating system does sector buffering with that type of IO.
//
// The other implementation is using fopen etc. which is buffered
// IO.  
//
// The implementation that you use depends on the second argument
// to the constructor - which is defaulted to the buffered IO (for
// now at least)
//
// One reason I have named this SERVbinaryFile is to avoid one of
// the classic mistakes with fopen which is to forget to put in the "b"
// to indicate that you wish to open the file in binary mode.  Without
// that flag the OS will helpfully translate newlines for you which
// results in some perplexing hard to find bugs.
//
// Notice that SERVbinaryFile implements SERVsink.
//
// The destructor of SERVbinaryFile will automatically close the file
// it may have open.
//
// SERVbinaryFile is a lot cleaner to use than using the fopen etc.
// calls directly since it gives a simpler interface, better error checking
// and hides some of nasty operating system implementation dependent
// details underneath.
//
// Date:   Fri 01/16/2004 
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVbinaryFile.h,v $
//---------------------------------------------------------------------------
#include "SERVsink.h"

class CHMstring;
class SERVbinaryFilePrivate;

typedef size_t SERVindex;
typedef CHMuint64 SERVfileIndex; // Allow large files.

class SERVbinaryFile 
: public SERVsink
{
public:
   enum SERVtype
   {
      Buffered,
      Unbuffered
   };

   enum SERVmode
   {
      // open for reading only
      Read,
      // open for reading and writing
      Write,
      // open and seek to the end
      Append,
      // open, truncate any existing file
      Rewrite
   };

   SERVbinaryFile
   (
      const CHMstring& FileName,
      SERVmode Mode,
      SERVtype Type=Buffered
   );

   SERVbinaryFile(SERVtype Type=Buffered);
   virtual ~SERVbinaryFile();

   virtual void open(const CHMstring& FileName,
                     SERVmode Mode);

   virtual void close();

   CHMboolean isOpen() const;

   virtual SERVindex position() const;
   virtual void setPosition(SERVindex NewPosition);

   // a pair of methods for positioning in files
   // with 64 bit addressing
   CHMuint64 position64() const;
   void setPosition64(CHMuint64 NewPosition);

   virtual CHMuint32 read(void* pBuffer, CHMuint32 SizeOfBuffer);

   // SERVsink interface method used to write the passed in data
   // to the sink object
   virtual CHMuint32 write(const void* pBuffer, CHMuint32 SizeOfBuffer);

   // Note that there is also a SERVfileSize function in SERVfileUtils.h.  That
   // works fine so long as the operating system does not have the file open.
   // However if you have the file open then only this method is guaranteed
   // to give the correct size of the file.
   SERVfileIndex size() const;

   const CHMstring& name() const;

   virtual void flush();

#ifdef WIN32
   //
   // The WIN32 HANDLE representing the file. 0 if the file has not been opened
   //
   void* win32Handle();
#endif

protected:
   CHMuint32 sourceEndPosition() const;
   void setName(const CHMstring& NewName);

private:
   SERVbinaryFilePrivate* pMember;

   SERVbinaryFile(const SERVbinaryFile& Orig); // not allowed
   SERVbinaryFile& operator=(const SERVbinaryFile& Orig); // not allowed
};

#endif // end of defensive include
