//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVhashTable
//
// Description:
//
// Templated hash table collection class.
//
// Date:   02/05/2004
// Revision: $Revision: 1.2 $
// Source:   $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVhashTable.cpp,v $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <CHM/CHMerrorClass.h>
#include <CHM/CHMstring.h>
#include "SERVhashTable.h"

const SERVindex SERV_MAX_CHAIN_LENGTH = 4; // when number of nodes in a bucket exceeds this, resize Buckets

void SERVvoidHashTable::init(void)
{
   // A series of prime numbers; must end with 0.
   static const SERVindex BucketArraySizes[] =
   {
      2, 5, 11, 17, 29, 47, 71, 107, 163, 251, 379, 569, 857, 1289,
      1949, 2927, 4391, 6599, 9901, 14867, 22303, 33457, 50207,
      75323, 112997, 169501, 254257, 381389, 572087, 858149, 0
   };

   Size = 0;
   Buckets = new SERVhashPlace[2];
   Buckets[0] = NULL;
   Buckets[1] = NULL;
   pCountOfBucket = &BucketArraySizes[0]; // == 2
}

SERVvoidHashTable::~SERVvoidHashTable(void)
{
   delete[] Buckets;
}

void SERVvoidHashTable::clear(void)
{
   SERVhashPlace CurrentPlace;
   const SERVindex CountOfBucket = *pCountOfBucket;
   for (SERVindex BucketIndex = 0; BucketIndex < CountOfBucket; BucketIndex++)
   {
      CurrentPlace = Buckets[BucketIndex];
      while (CurrentPlace != NULL)
      {
         SERVhashPlace Place = CurrentPlace;
         CurrentPlace = CurrentPlace->Next;
         destroyItem(Place);
      }
      Buckets[BucketIndex] = NULL;
   }
   try
   {
      SERVhashPlace* OldBuckets = Buckets;
      init();
      delete[] OldBuckets;
   }
   catch (...)
   {
      Size = 0;
      // suppress exception because it is not fatal - no memory
   }
   invariant();
}

void SERVvoidHashTable::remove(SERVhashPlace Place)
{
   CHMPRECONDITION(Place != NULL);
   if (Place == NULL)
   {
      return; // paranoid check
   }
   SERVindex BucketIndex = hashItem(Place) % *pCountOfBucket;
   if (Place == Buckets[BucketIndex])
   {
      Buckets[BucketIndex] = Place->Next;
   }
   else
   {
      for (SERVhashPlace Current = Buckets[BucketIndex]; Current != NULL; Current = Current->Next)
      {
         if (Current->Next == Place)
         {
            Current->Next = Place->Next;
            break;
         }
         CHMPRECONDITION(Current->Next != NULL);
      }
   }
   destroyItem(Place);
   Size--;
   invariant();
}

SERVhashPlace SERVvoidHashTable::first(void) const
{
   SERVhashPlace Current;
   const SERVindex CountOfBucket = *pCountOfBucket;
   for (SERVindex BucketIndex = 0; BucketIndex < CountOfBucket; BucketIndex++)
   {
      if ((Current = Buckets[BucketIndex]) != NULL)
      {
         return Current;
      }
   }
   return NULL;
}

SERVhashPlace SERVvoidHashTable::last(void) const
{
   for (SERVindex BucketIndex = *pCountOfBucket; BucketIndex > 0; BucketIndex--)
   {
      for (SERVhashPlace Current = Buckets[BucketIndex - 1]; Current != NULL; Current = Current->Next)
      {
         if (Current->Next == NULL)
         {
            return Current;
         }
      }
   }
   return NULL;
}

SERVhashPlace SERVvoidHashTable::previous(SERVhashPlace Place) const
{
   SERVindex BucketIndex = (Place == NULL ? *pCountOfBucket : hashItem(Place) % *pCountOfBucket);
   if (Place != NULL && Place != Buckets[BucketIndex])
   {
      for (SERVhashPlace Current = Buckets[BucketIndex]; Current != NULL; Current = Current->Next)
      {
         if (Current->Next == Place)
         {
            Place = Current;
            break;
         }
         CHMPRECONDITION(Current->Next != NULL);
      }
   }
   else
   {
      Place = NULL;
      for (; BucketIndex > 0; BucketIndex--)
      {
         if (Buckets[BucketIndex - 1] != NULL)
         {
            for (SERVhashPlace Current = Buckets[BucketIndex - 1]; Current != NULL; Current = Current->Next)
            {
               if (Current->Next == NULL)
               {
                  Place = Current;
                  break;
               }
            }
            break;
         }
      }
   }
   return Place;
}

SERVhashPlace SERVvoidHashTable::next(SERVhashPlace Place) const
{
   if (Place != NULL && Place->Next != NULL)
   {
      Place = Place->Next;
   }
   else
   {
      const SERVindex CountOfBucket = *pCountOfBucket;
      SERVindex BucketIndex = (Place == NULL ? (SERVindex)-1 : SERVindex(hashItem(Place) % CountOfBucket));
      Place = NULL;
      for (BucketIndex++; BucketIndex < CountOfBucket; BucketIndex++)
      {
         if ((Place = Buckets[BucketIndex]) != NULL)
         {
            break;
         }
      }
   }
   return Place;
}

CHMboolean SERVvoidHashTable::removeItem(SERVindex Hash, const void* Key)
{
   SERVhashPlace Current = findItem(Hash, Key);
   if (Current != NULL)
   {
      remove(Current);
      return true;
   }
   return false;
}

inline void SERVvoidHashTable::grow(void)
{
   const SERVindex NewCountOfBucket = *(pCountOfBucket+1);
   SERVhashPlace* NewBuckets = 0;
   
   if( NewCountOfBucket )
   {
      try
      {
         NewBuckets = new SERVhashPlace[NewCountOfBucket];
      }
      catch (...)
      {
         // Not enough memory, so we won't increase the number of buckets.
      }
   }

   if( NewBuckets )
   {
      SERVindex CountOfBucket = *pCountOfBucket;

      SERVhashPlace Chain = NULL;
      SERVhashPlace Current, Place;
      SERVindex       Bucket;

      // Insert all items into a temporary Chain

      for (Bucket = 0; Bucket < CountOfBucket; Bucket++)
      {
         Current = Buckets[Bucket];
         while (Current != NULL)
         {
            Place = Current;
            Current = Current->Next;
            Place->Next = Chain;
            Chain = Place;
         }
      }

      // Switch over to the new, larger bucket array.

      delete[] Buckets;
      Buckets = NewBuckets;
      ++pCountOfBucket;

      CountOfBucket = NewCountOfBucket;

      // Set all bucket pointers to NULL

      for (Bucket = 0; Bucket < CountOfBucket; Bucket++)
      {
         Buckets[Bucket] = NULL;
      }

      // Reinsert all the existing items

      while (Chain != NULL)
      {
         Place = Chain;
         Chain = Chain->Next;
         Bucket = hashItem(Place) % CountOfBucket;
         Place->Next = Buckets[Bucket];
         Buckets[Bucket] = Place;
      }
   }
}

SERVhashPlace SERVvoidHashTable::addItem(SERVindex Hash, const void* Key, SERVhashPlace Node)
{
   if (Node == NULL)
   {
      SERV_THROW_ERROR("Out of memory", 0);
   }

   SERVindex CountOfBucket = *pCountOfBucket;

   SERVindex BucketIndex = Hash % CountOfBucket;
   SERVindex ChainLength = 0;

   // DiagCheck for an existing Item that matches the Key

   for (SERVhashPlace Place = Buckets[BucketIndex]; Place != NULL; Place = Place->Next)
   {
      if (equalKey(Key, Place))
      {
         destroyItem(Node); // clean up the duplicate Item
         return NULL;
      }
      ChainLength++;
   }

   // If the chain length is getting long then upsize the bucket array

   if (ChainLength >= SERV_MAX_CHAIN_LENGTH && Size >= CountOfBucket)
   {
      grow();
      CountOfBucket = *pCountOfBucket;
      BucketIndex = Hash % CountOfBucket;
   }

   // Insert the new Item

   Node->Next = Buckets[BucketIndex];
   Buckets[BucketIndex] = Node;
   Size++;
   invariant();
   return Node;
}

SERVhashPlace SERVvoidHashTable::findItem(SERVindex Hash, const void* Key) const
{
   SERVindex BucketIndex = Hash % *pCountOfBucket;
   for (SERVhashPlace Current = Buckets[BucketIndex]; Current != NULL; Current = Current->Next)
   {
      if (equalKey(Key, Current))
      {
         return Current;
      }
   }
   return NULL;
}

#ifdef DEBUG
void SERVvoidHashTable::invariant(void) const
{
   const SERVindex CountOfBucket = *pCountOfBucket;

   SERVindex NodeCount = 0;
   for (SERVindex BucketIndex = 0; BucketIndex < CountOfBucket; BucketIndex++)
   {
      for (SERVhashPlace Current = Buckets[BucketIndex]; Current != NULL; Current = Current->Next)
      {
         Assert(hashItem(Current) % CountOfBucket == (SERVindex) BucketIndex); // Item is in correct bucket
         NodeCount++;
      }
   }
   Assert(Size == NodeCount);
}
#endif
