#ifndef __SERV_HASH_TABLE_H__
#define __SERV_HASH_TABLE_H__
//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVhashTable
//
// Description:
//
// Hash table collection class.
//
// When a collection is constructed it is passed a hash function that
// returns a SERVindex.
//
// SERVindex HashProc(const KeyType& Key);
//
// The operator == is used to test that keys match.
//
// The objects stored in the template must support the copy contructor
// Type(const Type&)
//
// A copy is made of an item when it is inserted into the list. When
// an item is removed from the list it is destructed.
//
// To make a list which stores references or pointers declare the list
// so that the item type is a pointer or a reference. Iterating through
// a list is done using SERVhashTablePlace variables.
//
// SERVhashTable<CHMstring, CustomerRecord*> MyHashTable(HashString);
// SERVhashTablePlace                        Place;
//
// for (Place = MyHashTable.first(); Place != NULL; Place = MyHashTable.next(Place))
// {
//    printCustomer(MyHashTable[Place]);
// }
// delete MyHashTable[MyHashTable.first()];
// MyHashTable.remove(MyHashTable.first());
// MyHashTable.add("unnamed", new CustomerRecord("Unnamed"));
//
// Passing a NULL to methods that take a Place argument is handled in the
// following ways
//
// previous(NULL) returns the last item
// next(NULL) returns the first item
//
// Notice the usage of the SERVvoidHashTable.  This is used in support of the thin
// template pattern for faster compile time.  See intranet for more details.
//
// Date:   02/05/2004
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVhashTable.h,v $
//---------------------------------------------------------------------------

#include <CHM/CHMerrorClass.h>
#include "SERVdebugLog.h"
#include "SERVhash.h"

class SERVhashNode
{
public:
   SERVhashNode* Next;
};

typedef SERVhashNode* SERVhashPlace;

class SERVvoidHashTable
{
private:
   SERVhashPlace*   Buckets;
   const SERVindex*   pCountOfBucket;
   SERVindex          Size;
   void init(void);
   inline void grow(void);
public:
   SERVvoidHashTable(void) { init(); }
   virtual ~SERVvoidHashTable(void);

   SERVindex size(void) const { return Size; }
   void clear(void);
   void remove(SERVhashPlace Place);

   SERVhashPlace first(void) const;
   SERVhashPlace last(void) const;
   SERVhashPlace previous(SERVhashPlace Place) const;
   SERVhashPlace next(SERVhashPlace Place) const;
protected:
   CHMboolean removeItem(SERVindex Hash, const void* Key);
   SERVhashPlace addItem(SERVindex Hash, const void* Key, SERVhashPlace Node);
   SERVhashPlace findItem(SERVindex Hash, const void* Key) const;
protected:
   virtual void destroyItem(SERVhashPlace Place) = 0;
   virtual SERVindex hashItem(SERVhashPlace Place) const = 0;
   virtual CHMboolean equalKey(const void* Key, SERVhashPlace Place) const = 0;
private:
   SERVvoidHashTable(const SERVvoidHashTable& Lookup);               // disable copy constructor
   SERVvoidHashTable& operator =(const SERVvoidHashTable& Lookup);   // disable assignment
#ifdef DEBUG
   void invariant(void) const;
#else
   void invariant(void) const {}
#endif
};

typedef SERVhashPlace SERVhashTablePlace;

template <class KeyType, class Type>
struct SERVhashTableNode: public SERVhashNode
{
   SERVindex Hash;
   KeyType  Key;
   Type     Item;

   SERVhashTableNode(SERVindex iHash, 
          const KeyType& iKey, 
          const Type& iItem)
      : Hash(iHash), 
        Key(iKey), 
   Item(iItem) {}
};

template <class KeyType, class Type, class CompareType = SERVhash<KeyType> >
class SERVhashTable: public SERVvoidHashTable
{
public:
   //public typedefs for generic functions
   typedef KeyType            TKey;
   typedef Type               TValue;
   typedef SERVhashTablePlace TPlace;

   typedef SERVindex (*HashProc)(const KeyType& Key);

   SERVhashTable(): Hash(CompareType::defaultHash) {}
   SERVhashTable(HashProc TheHashProc): Hash(TheHashProc) {}

   virtual ~SERVhashTable(void)
   { 
      try
      {
         clear(); 
      }
      catch (const CHMerror& Error)
      {
         SERV_LOG(Error);
      }
   }

   Type& operator [](SERVhashTablePlace Place)
   {
      CHMPRECONDITION(Place != NULL);
      return ((SERVhashTableNode<KeyType, Type>*) Place)->Item;
   }

   const Type& operator [](SERVhashTablePlace Place) const
   {
      CHMPRECONDITION(Place != NULL);
      return ((SERVhashTableNode<KeyType, Type>*) Place)->Item;
   }
   
   KeyType& key(SERVhashTablePlace Place)
   {
      CHMPRECONDITION(Place != NULL);
      return ((SERVhashTableNode<KeyType, Type>*) Place)->Key;
   }

   const KeyType& key(SERVhashTablePlace Place) const
   {
      CHMPRECONDITION(Place != NULL);
      return ((SERVhashTableNode<KeyType, Type>*) Place)->Key;
   }

   Type& operator [](const KeyType& Key);
   const Type& operator [](const KeyType& Key) const;

   CHMboolean removeKey(const KeyType& Key) { return removeItem(Hash(Key), &Key); }
   SERVhashTablePlace add(const KeyType& Key, const Type& Item);
   SERVhashTablePlace find(const KeyType& Key) const { return findItem(Hash(Key), &Key); }

   template<class TClass>
   void iterate(void (TClass::*pMethodVoid)(const KeyType&,Type&),TClass* pInstance, SERVhashPlace From = NULL, SERVhashPlace To = NULL)
   {
      if (!From)
      {
         From = this->first();
      }
      while (From != NULL)
      {
         (pInstance->*pMethodVoid)(this->key(From),this->operator[](From));
         if (To && To == From)
         {
            break;
         }
         From = this->next(From);
      }
   }

   template<class TClass>
   SERVhashPlace iterate(SERVhashPlace (TClass::*pMethodVoid)(const KeyType&,Type&,SERVhashPlace CurrentPlace),TClass* pInstance, SERVhashPlace From=NULL, SERVhashPlace To=NULL)
   {
      if (!From)
      {
         From = this->first();
      }
      while (From != NULL)
      {
         SERVhashPlace Found = (pInstance->*pMethodVoid)(this->key(From),this->operator[](From),From);
         if (Found)
         {
            return Found;
         }
         if (To && To == From)
         {
            break;
         }
         From = this->next(From);
      }
      return NULL;
   }

protected:
   virtual void destroyItem(SERVhashTablePlace Place) { delete (SERVhashTableNode<KeyType, Type>*) Place; }
   virtual SERVindex hashItem(SERVhashTablePlace Place) const { return ((SERVhashTableNode<KeyType, Type>*) Place)->Hash; }
   virtual CHMboolean equalKey(const void* Key, SERVhashTablePlace Place) const 
   {
      return CompareType::equalKey(*(KeyType*) Key,((SERVhashTableNode<KeyType, Type>*) Place)->Key);
   }

private:
   HashProc    Hash;
};

template <class KeyType, class Type, class CompareType>
Type& SERVhashTable<KeyType, Type, CompareType>::operator [](const KeyType& Key)
{
   SERVhashTablePlace Place = find(Key);
   if (Place == NULL)
   {
      SERVindex HashValue = Hash(Key);
      Place = addItem(HashValue, &Key, new SERVhashTableNode<KeyType, Type>(HashValue, Key, Type()));
   }
   return ((SERVhashTableNode<KeyType, Type>*) Place)->Item;
}

template <class KeyType, class Type, class CompareType>
const Type& SERVhashTable<KeyType, Type, CompareType>::operator [](const KeyType& Key) const
{
   SERVhashTablePlace Place = find(Key);
   CHMPRECONDITION(Place != NULL);
   return ((SERVhashTableNode<KeyType, Type>*) Place)->Item;
}

template <class KeyType, class Type, class CompareType>
SERVhashTablePlace SERVhashTable<KeyType, Type, CompareType>::add(const KeyType& Key, const Type& Item)
{
   SERVindex HashValue = Hash(Key);
   return addItem(HashValue, &Key, new SERVhashTableNode<KeyType, Type>(HashValue, Key, Item));
}

#endif // end of defensive include
