#ifndef __SERV_OSTREAM_H__
#define __SERV_OSTREAM_H__
//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVostream
//
// Description:
//
// SERVostream translates C and user defined types into text.
//
// Each SERVostream object has a SERVsink object which is used to
// write the actual formatted text.  Each SERVstream object may or
// may not 'own' the SERVsink object that it writes to.
//
// There are a number of standard global objects SERVcout, SERVcerr,
// SERVclog which are commented on further down this file.
//
// To format numbers in hexidecimal we can do the following:
//
// SERVcout << hex << Number << " (in hex) " << decimal << Number << 
//        " (in decimal) " << newline << flush;
//
// Note the use of the newline and flush operators.  These will
// add a newline and flush the sink respectively.  Since different
// operating systems have different conventions for newlines there is
// a setNewLine method which allows one to specify the string to insert
// for a newline.
//
// Sometimes I have found it necessary to write:
//
// SERVcout << ::flush;  
//
// The problem here is when the above statement is in the context of
// a class which actually implements a flush method - basically anything
// which implements SERVsink has this problem.
//
// SERVostream can be useful way to convert numbers into strings i.e.
//
// CHMuint32 MyNumber = 1221;
// CHMstring MyString;
// SERVostream Stream(MyString);
//
// Stream << MyNumber;
//
// Date:   Fri 01/16/2000 
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVostream.h,v $
//---------------------------------------------------------------------------

#include <CHM/CHMminimumInclude.h>

class CHMstring;
class SERVsink;
class SERVostreamPrivate;

class SERVostream
{
public:
   // This constructor takes ownership of the sink object.
   SERVostream(SERVsink* Sink, CHMboolean Owner);
   // This constructor does not take ownership of the sink object.
   SERVostream(SERVsink& Sink);

   virtual ~SERVostream();

   void flush();
   // The next few functions don't quite conform to our naming conventions - bit
   // of legacy - oh well.
   void newline();

   void setNewLine(const CHMchar* NewLine);
   const CHMchar* getNewLine() const;

   // The indent content refers to what will be used to make an indented
   // entry.  By default this is set to "   " but one might for certain applications
   // set this to "\t" for a more space efficient indenting mechanism.
   // Most of the time it will be more elegant to use the overloaded operators
   // to invoke these indentation methods i.e.
   // Stream << set_indent("\t")
   //        << add_indent
   //        << "Blah"
   //        << remove_indent;
   // When one wants to remove all indentation one can do:
   // Stream << clear_indent;
   void setIndentContent(const CHMstring& Indent);
   const CHMstring& indentContent() const;

   void indent();
   void unindent();
   void resetIndent();
   void setCountOfIndent(CHMuint32 NewValue);
   CHMuint32 countOfIndent() const;

   virtual void setfill(CHMuint32 n);

   SERVostream& operator<<(SERVostream&(*)(SERVostream&)); 
   SERVostream& operator<<(SERVostream&(*)(SERVostream&, int n)); 
   SERVostream& operator<<(CHMchar Character);
   SERVostream& operator<<(unsigned char Character);
   SERVostream& operator<<(const CHMchar* pString);

   SERVostream& operator<<(const CHMstring& String);

   SERVostream& operator<<(CHMfloat64 Number);
   SERVostream& operator<<(CHMfloat32 Number);

   SERVostream& operator<<(CHMint16 Number);
   SERVostream& operator<<(CHMuint16 Number);
   
   SERVostream& operator<<(CHMint32 Number);
   SERVostream& operator<<(CHMuint32 Number);
   
   SERVostream& operator<<(CHMuint64 Number);   
   SERVostream& operator<<(CHMint64 Number);   

   //modifiers are redirected
   SERVostream& operator<<(unsigned long Number)
   {
      return operator<< ((CHMuint32)Number);
   }   

   SERVostream& operator<<(long Number)
   {
      return operator<< ((CHMint32)Number);
   }   

#ifndef CHM_NO_BOOLEAN_TYPE
   // Just to make life marginally easier, let's define a SERVostream that prints CHMboolean 
   // values as "true" or "false"
   SERVostream& operator<<(CHMboolean Boolean);
#endif

   // returns true if this SERVostream owns its sink object.
   CHMboolean isSinkOwner();
   SERVsink* sink();
   
   SERVsink* setSink(SERVsink* pSink, CHMboolean IsOwner);

   SERVsink& setSink(SERVsink& Sink)
   {
      return *setSink(&Sink, false);
   }

private:
   SERVostreamPrivate* pMember;

   enum ENumberBase
   {
      Decimal = 10,
      Hexadecimal = 16
   };
   ENumberBase numberBase();
   void setNumberBase( ENumberBase Base );

   friend SERVostream& hex(SERVostream& _outs);

   friend SERVostream& decimal(SERVostream& _outs);

   friend class SERVostreamPrivate;

   SERVostream(const SERVostream& Orig);
   SERVostream& operator=(const SERVostream& Orig);
};

// SERVcerr, SERVcout and SERVclog are equivalent to the
// the standard iostream library's cerr, cout, clog 
// objects.  They are global SERVostream objects that go
// to standard error, standard out and in the case of
// SERVclog usually standard out.  
//
// It is a common technique in our code to re-direct
// these objects to point to different sink objects - i.e.
// we may choose to redirect to a file using SERVbinaryFile.

extern SERVostream SERVcerr;
extern SERVostream SERVcout;
extern SERVostream SERVclog;

inline SERVostream& flush(SERVostream& _outs)
{
   _outs.flush(); return _outs;
}

inline SERVostream& newline(SERVostream& _outs)
{
   _outs.newline(); return _outs; 
}

SERVostream& indent(SERVostream& OutStream);
SERVostream& unindent(SERVostream& OutStream);
SERVostream& reset_indent(SERVostream& OutStream);

SERVostream& hex(SERVostream& _outs);

SERVostream& decimal(SERVostream& _outs);

class SERVostreamManip
{
public:
   virtual void printOn(SERVostream& Stream) const {}

   virtual void copy() {}
};

// This complicated looking stuff is for implementation SERVostream manipulators
// with arguments.
template<class TArgument>
class SERVostreamManip1 : public SERVostreamManip
{
public:
   typedef void (*TSERVostreamFunction)(SERVostream&, const TArgument&);
   SERVostreamManip1(TSERVostreamFunction ipFunction, const TArgument& iArgument)
   : pFunction(ipFunction),
     Argument(iArgument)
   {
   }

   virtual void printOn(SERVostream& Stream) const
   {
      (*pFunction)(Stream, Argument);
   }

   TSERVostreamFunction pFunction;
   TArgument Argument;
};

inline SERVostream& operator<<(SERVostream& Stream, SERVostreamManip* pManipulator)
{
   pManipulator->printOn(Stream);
   delete pManipulator;
   return Stream;
}

SERVostreamManip* set_indent(const CHMstring& TagName);

SERVostreamManip* setfill(CHMchar FillChar);

#endif // end of defensive include
