//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVserver
//
// Description:
//
// Implementation
//
// Date:   Wed 05/17/2007
// Revision: $Revision: 1.4 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVserver.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <signal.h>
#include <stdlib.h>
#include <windows.h>
#include <CHM/CHMstring.h>
#include <CHM/CHMerrorClass.h>
#include "SERVostream.h"
#include "SERVserver.h"
#include "SERVutils.h"

class SERVserverPrivate
{
public:
   SERVserverPrivate(const CHMstring& iServiceName, SERVserver* ipParent)
   : ServiceName(iServiceName),
     HasStarted(false)
   {
      SERV_EXITCONDITION(pThis == NULL); // There can be only one singleton!

      pParent = ipParent;
      pThis = this;
   }

   ~SERVserverPrivate(){}

   // The purpose of pParent and pThis is so that we can access object members 
   // from static functions.
   static SERVserverPrivate* pThis;
   SERVserver* pParent;

   static void initializeService();

   static void WINAPI serviceMain(DWORD argc, LPTSTR* argv);
   static void WINAPI handler(DWORD OperationCode);
   
   // Signal handler for a Ctrl + C signal.
   static void __cdecl ControlCHandler(int SignalValue);

   void setServerStatus(DWORD NewStatus);
   DWORD serviceStatus() const;

   void initControlCSignalHandler();

   void notifyStartProblem(CHMuint32 ErrorCode);

   SERVICE_STATUS_HANDLE ServiceStatusHandle;
   SERVICE_STATUS ServiceStatus;
   CHMboolean HasStarted;
   CHMstring ServiceName;
};

SERVserverPrivate* SERVserverPrivate::pThis = NULL; 

// Standard initialization of status that is required by windows
// In ControlsAccepted we have restricted accepted codes to stop and implicit interrogate
// Thus, no PAUSE_CONTINUE or SHUTDOWN
void SERVserverPrivate::initializeService()
{  
   SERVserverPrivate* pServer = SERVserverPrivate::pThis;
   CHMPRECONDITION(pServer != NULL);

   pServer->ServiceStatus.dwServiceType        = SERVICE_WIN32; // Service type
   pServer->ServiceStatus.dwCurrentState       = SERVICE_START_PENDING; // Current status
   pServer->ServiceStatus.dwControlsAccepted   = SERVICE_ACCEPT_STOP;  // Accepted control codes                                                
   pServer->ServiceStatus.dwWin32ExitCode      = 0; // Error code
   pServer->ServiceStatus.dwServiceSpecificExitCode = 0; // Specific error code - when starting or stopping
   pServer->ServiceStatus.dwCheckPoint         = 0; // Value that the service increments to show status in operations such as start,stop...
   pServer->ServiceStatus.dwWaitHint           = 0; // Estimate amount of time while completing operation before next call
}

void SERVserverPrivate::serviceMain(DWORD argc, LPTSTR* argv)
{
   SERVserverPrivate* pServer = SERVserverPrivate::pThis; 
   SERV_EXITCONDITION(pServer != NULL);

   // Handler function is registered. i.e. If a control request comes in, it will be handled by 
   // the handler function. 
   pServer->ServiceStatusHandle = ::RegisterServiceCtrlHandler(pServer->ServiceName.c_str(), // address of name of service
                                                               SERVserverPrivate::handler); // address of handler function
     
   if (0 == pServer->ServiceStatusHandle)
   {
      SERV_THROW_WIN32_ERROR("trying to register handle of server in SERVserverPrivate::serviceMain()");
   }

   pServer->initializeService();
   pServer->setServerStatus(SERVICE_START_PENDING);
   pServer->HasStarted = true;

   if (pServer->pParent->onInit(true))
   {
      pServer->setServerStatus(SERVICE_RUNNING); // Set the status before asking the service to run.
      pServer->pParent->onRun(); // NOTE: onRun() will keep executing until it gets a signal to stop.
   }

   pServer->setServerStatus(SERVICE_STOPPED);
}

void SERVserverPrivate::handler(DWORD OperationCode)
{
   // Get a pointer to the object
   SERVserverPrivate* pServer = SERVserverPrivate::pThis;
   CHMPRECONDITION(pServer != NULL);

   switch (OperationCode) 
   {
   case SERVICE_CONTROL_STOP:
      pServer->setServerStatus(SERVICE_STOP_PENDING);        
      pServer->pParent->onStop(); // server would be running until this is signaled
      break;
   case SERVICE_CONTROL_INTERROGATE:
      pServer->setServerStatus(pServer->serviceStatus());
      break;
   }
}

// Signal Handler
void __cdecl SERVserverPrivate::ControlCHandler(int SignalValue)
{
   CHMPRECONDITION(SERVserverPrivate::pThis != NULL);
   CHMPRECONDITION(SERVserverPrivate::pThis->pParent != NULL);
   SERVcout << "Ctrl+C detected, sending stop signal." << newline << flush; 
   pThis->pParent->onStop();
}

void SERVserverPrivate::setServerStatus(DWORD NewStatus)
{
   CHMPRECONDITION(ServiceStatusHandle != NULL)
   ServiceStatus.dwCurrentState  = NewStatus;
   if (0 == ::SetServiceStatus(ServiceStatusHandle, &ServiceStatus))
   {
      SERV_THROW_WIN32_ERROR("trying to set status of service")
   }
}

DWORD SERVserverPrivate::serviceStatus() const
{
   CHMPRECONDITION(ServiceStatusHandle != NULL)
   return ServiceStatus.dwCurrentState;
}

void SERVserverPrivate::initControlCSignalHandler()
{
   if (SIG_ERR == ::signal(SIGINT, &SERVserverPrivate::ControlCHandler))
   {
      SERV_THROW_WIN32_ERROR("trying to set interrupt signal handling (Ctrl+c) in SERVserver::runServerAsNormalProcess()");
   }
} 

void SERVserverPrivate::notifyStartProblem(CHMuint32 ErrorCode)
{
   CHMPRECONDITION(ServiceStatusHandle != NULL);
   ServiceStatus.dwCurrentState = SERVICE_STOPPED;
   ServiceStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
   ServiceStatus.dwServiceSpecificExitCode = ErrorCode;

   if (0 == ::SetServiceStatus(ServiceStatusHandle, &ServiceStatus))
   {
      SERV_THROW_WIN32_ERROR("trying to notify start problem")
   }
}

SERVserver::SERVserver(const CHMstring& ServiceName)
 : pMember(NULL)
{
   pMember = new SERVserverPrivate(ServiceName, this);
}

SERVserver::~SERVserver()
{
   pMember->pParent = NULL;
   delete pMember;
}

const CHMstring& SERVserver::name() const
{
   return pMember->ServiceName;
}

// We invoke this method to start off the service
CHMboolean SERVserver::runServer()
{
   SERVICE_TABLE_ENTRY ServiceTable[] = 
   {
      {pMember->ServiceName.get_buffer() , SERVserverPrivate::serviceMain},
      {NULL, NULL}
   };

   if (0 == ::StartServiceCtrlDispatcher(ServiceTable))
   {
      DWORD ErrorCode = ::GetLastError();

      // Dispatcher does not succeed however error is due to the fact that service is run from command line
      if (ERROR_FAILED_SERVICE_CONTROLLER_CONNECT == ErrorCode)
      {  
         return false;
      }
      else
      {
         SERV_THROW_WIN32_ERROR("trying to start service server in SERVserver::runService()");
      }
   }
   return true;
}

void SERVserver::runServerAsNormalProcess()
{
   SERVcout << "Running service as command-line application.  Press Ctrl+C to exit." << newline << newline << flush;

   // Specific signal and its handler are specified. The code moves on after that.
   pMember->initControlCSignalHandler();
   if (onInit(false))
   {
      pMember->HasStarted = true;
      onRun();
   }

   SERVcout << "Exiting server." << newline << flush;
}

void SERVserver::notifyStartProblem(CHMuint32 ErrorCode)
{
   pMember->notifyStartProblem(ErrorCode);
}
