//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: C++ Server Service Example
//
// Description:
//
// Implementation
//
// Date:   Wed 05/17/2007
// Revision: $Revision: 1.5 $
//
// Last Edit Date: $Date: 2007/06/20 16:21:21 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/ServerService/CHMserver.cpp,v $
//---------------------------------------------------------------------------
#include "CHMserverServicePrecomp.h"
#pragma hdrstop

#include <CHM/CHMllpConnection.h>
#include <BIZ/BIZutils.h>
#include <SERV/SERVostream.h>
#include <SERV/SERVutils.h>
#include "CHMserver.h"
#include "CHMconstants.h"

static const CHMint32 MAX_MESSAGE_LENGTH = 200;

CHMserver::CHMserver()
{
   // We initialize the engine used for auto-acknowledgment/negative acknowledgment.
   // We use these engines to generate HL7 acknowledgment (ACK) and negative 
   // acknowledgment (NACK) messages.  Here the Engines we use are DynamicEngine.
   CHMstring AppFilePath;
   SERVgetExecutableFileDirectory(AppFilePath);

   AutoAckEngine.Load(CHMstring(AppFilePath + "autoack.vmd").c_str());
   AutoNackEngine.Load(CHMstring(AppFilePath + "autonack.vmd").c_str());
}

// This is the event we will get when a Client is connected to the Server.
// The 'Client' object (of type CHMllpConnection) can be used to interact with the machine that is sending
// the data.  There can be multiple clients connected to the server.
void CHMserver::OnClientConnected(CHMllpConnection& Client)
{
   // TODO - Catch C++ exceptions or assume SERVcout doesn't throw.
   SERVcout << "Accepted Connection From " << Client.RemoteIpAddressAsString() << newline
            << "Current number of Client Connection is " << this->CountOfClient() << newline << flush;
}

// This is the event we will get when a Client disconnects from the server. Note that sometimes
// if the client is not well behaved we may not receive this event.
void CHMserver::OnClientDisconnected(CHMllpConnection& Client)
{
   // TODO - Catch C++ exceptions or assume SERVcout doesn't throw.
   SERVcout << "Disconnected From " << Client.RemoteIpAddressAsString() << newline << flush;
}

// This event is the main entry point when we get an HL7 message from a connected 
// peer.
//
// Note that the LLP TCP/IP transport components are not tightly coupled to the
// hl7Engine object.  You can use any kind of transport you like with Chameleon, including
// your own transports or third party components.
//
// Notice that in this application we do not block on events - this is because the
// networking components are asynchronous.  You can write synchronous applications using
// Chameleon but you would need to use a transport mechanism other than the LLP TCP/IP
// components supplied here.
void CHMserver::OnMessageReceived(CHMllpConnection& Client,const char* Message)
{
   try
   {
      SERVcout << "Received a message:" << newline
               << MakeMessagePrintable(Message, APP_NEW_LINE, MAX_MESSAGE_LENGTH) << newline << flush;

      // Process the message here.
      BIZprocessMessage(Message, SERVcout);

      sendAckToClient(Client, Message);
   }
   catch (const CHMerror& Error)
   {
      SERVcerr << "ERROR: OnMessageReceived" << newline << Error.Description() << newline << flush;
   }
   catch (...)
   {
      SERVcerr << "ERROR: OnMessageReceived" << newline << "Unknown exception." << newline << flush;
   }
}

// This is the event we'll get when we have an error on a given TCP/IP connection.
void CHMserver::OnClientErrorReceived(CHMllpConnection& Client, const char* ErrorMessage, CHMint32 ErrorCode)
{
   // TODO - Catch C++ exceptions or assume SERVcout doesn't throw.
   SERVcout << "Error received from Client side (IP: " << Client.RemoteIpAddressAsString() << "):" << newline
            << ErrorMessage << newline << flush;
}

// This event fires if we have an error related to the listening socket itself rather
// than a specific connection.  In practice this event is rarely invoked.
void CHMserver::OnErrorReceived(const char* ErrorMessage, CHMint32 Code)
{
   // TODO - Catch C++ exceptions or assume SERVcout doesn't throw.
   SERVcout << "Socket Error:" << newline
            << ErrorMessage << newline << flush;

   // Disconnect all clients on server error.
   DisconnectAllClients();
}

// This event fires when we receive data that is not using the Lower Layer Protocol (LLP).
// LLP is the standard protocol for sending HL7 messages, and is used by Chameleon.
// See the manual for more details.
void CHMserver::OnDataIgnored(CHMllpConnection& Client, const char* Data)
{
   // TODO - Catch C++ exceptions or assume SERVcout doesn't throw.
   SERVcout << "Received Ignored Data:" << newline
            << MakeMessagePrintable(Data, APP_NEW_LINE, MAX_MESSAGE_LENGTH) << newline << flush;
}

void CHMserver::sendAckToClient(CHMllpConnection& Client, const char* Message)
{
   const char* pAck;

   try
   {
      SERVcout << newline;

      // We are using auto-acknowledgment - see the manual for information.
      // Now we need to generate the ACK and send it back to the user.
      // We use auto-acknowledgment; to obtain more information regarding auto-acknowledgment,
      // please read the Chameleon manual.

      // This method generates an ACK using the ackEngine.
      pAck = AutoAckEngine.generateAck(Message);

      // We use the MakeMessagePrintable function to 'pretty print' the HL7 message 
      // into a format that can be easily read in the log file.  This function converts 
      // carriage return (\r) characters into the string given in the second argument of 
      // MakeMessagePrintable.  In this case it is converted into a carriage return/line feed 
      // (\r\n) sequence which displays nicely in the log file.
      // 
      // The third argument is the maximum length of the message before we truncate it.  
      // Having the third argument set to -1 means we are not truncating the output.
      SERVcout << "Ack generated:" << newline
               << MakeMessagePrintable(pAck, APP_NEW_LINE, MAX_MESSAGE_LENGTH) << newline
               << "Sending Ack back to Client." << newline << flush;
   }
   catch (CHMerror& Error)
   {
      SERVcerr << "An error occured while generating auto acknowledgement." << newline
               << Error.Description() << newline
               << "We will send a NACK instead." << newline << flush;

      try
      {
         // An error occurred! Send back a negative-acknowledgment.
         pAck = AutoNackEngine.generateAck(Message);
      }
      catch (CHMerror& Error)
      {
         // Just in case generating the NACK fails, we'll send a hard-coded NACK instead.
         SERVcerr << "Generating Nack Failed! We'll send a hard-coded nack instead." << newline << flush;

         pAck = "MSH|^~\\&|ACME|ACME_APP|XYZ|XYZ_APP|||NACK|XXX|P|2.4|\rMSA|AA|XXX|\r";
      }
   }
   catch (...)
   {
      SERVcerr << "Unhandled exception during auto acknowledgment." << newline << flush;
   }

   Client.TransmitMessage(pAck);
}
